"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
const tsutils_1 = require("tsutils");
const typescript_1 = __importDefault(require("typescript"));
const common_1 = require("./common");
/**
 * An ESLint rule to ban unsafe type assertions.
 */
const noUnsafeTypeAssertion = {
    meta: {
        type: "problem",
        docs: {
            category: "Possible Errors",
            description: "Bans unsafe type assertions.",
            recommended: "error",
            url: "https://github.com/danielnixon/eslint-plugin-total-functions",
        },
        messages: {
            errorStringGeneric: "This type assertion is not type-safe.",
            errorStringPreferAs: "Use the 'as' syntax for type assertions.",
        },
        schema: [],
    },
    create: (context) => {
        const parserServices = experimental_utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        const isUnsafe = (rawDestinationType, rawSourceType, seenTypes = []) => {
            const { destinationType, sourceType } = common_1.filterTypes(rawDestinationType, rawSourceType);
            // eslint-disable-next-line functional/no-conditional-statement
            if (destinationType === undefined || sourceType === undefined) {
                return false;
            }
            return destinationType.getProperties().some((destinationProperty) => {
                const destinationPropertyType = common_1.symbolToType(destinationProperty, checker);
                const sourceProperty = sourceType.getProperty(destinationProperty.name);
                const sourcePropertyType = sourceProperty !== undefined
                    ? common_1.symbolToType(sourceProperty, checker)
                    : undefined;
                const destinationPropertyIsOptional = destinationProperty.flags & typescript_1.default.SymbolFlags.Optional;
                const destinationTypeParts = destinationPropertyType !== undefined
                    ? tsutils_1.unionTypeParts(destinationPropertyType)
                    : [];
                const destinationPropertyAllowsUndefined = destinationTypeParts.some((t) => t.flags & typescript_1.default.TypeFlags.Undefined);
                const sourcePropertyIsUndefined = sourcePropertyType === undefined ||
                    tsutils_1.unionTypeParts(sourcePropertyType).some((t) => t.flags & typescript_1.default.TypeFlags.Undefined);
                const nextSeenTypes = seenTypes.concat([
                    { destinationType, sourceType },
                ]);
                const isUnsafePropertyAssignment = sourcePropertyIsUndefined &&
                    !destinationPropertyAllowsUndefined &&
                    !destinationPropertyIsOptional;
                // This is an unsafe assignment if...
                return (
                // we're not in an infinitely recursive type,
                seenTypes.every((t) => t.destinationType !== destinationType &&
                    t.sourceType !== sourceType) &&
                    // and this is an unsafe property assignment, or
                    (isUnsafePropertyAssignment ||
                        // this is unsafe recursively
                        (destinationPropertyType !== undefined &&
                            sourcePropertyType !== undefined &&
                            isUnsafe(destinationPropertyType, sourcePropertyType, nextSeenTypes))));
            });
        };
        const reportUnsafe = (node
        // eslint-disable-next-line functional/no-return-void
        ) => {
            // The right hand side of the "as".
            const destinationNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            const rawDestinationType = checker.getTypeAtLocation(destinationNode);
            // The left hand side of the "as".
            const sourceNode = destinationNode.expression;
            const rawSourceType = checker.getTypeAtLocation(sourceNode);
            // eslint-disable-next-line functional/no-conditional-statement
            if (isUnsafe(rawDestinationType, rawSourceType)) {
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node,
                    messageId: "errorStringGeneric",
                });
            }
        };
        return {
            TSTypeAssertion: reportUnsafe,
            TSAsExpression: reportUnsafe,
        };
    },
};
exports.default = noUnsafeTypeAssertion;
